##############################################################################
#
# Enfold Stage Deployment - Remote Deployment of Staged Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: adapter.py 3551 2005-12-05 18:33:32Z sidnei $
"""

import time
from sets import Set
from types import ListType, TupleType
from entransit.client import Client
from zope.interface import implements
from DateTime import DateTime
from Acquisition import aq_inner, aq_parent, aq_base
from Products.StageDeployment.config import PICKLE_DEBUG
from Products.StageDeployment.formatters import default as default_formatter
from Products.StageDeployment.interface import IInformationExtractor
from Products.StageDeployment.interface import IReferenceExtractor
from Products.StageDeployment.interface import ILocalRolesExtractor
from Products.StageDeployment.interface import IDeploymentHandler
from Products.Archetypes.config import REFERENCE_CATALOG
from Products.Archetypes.utils import shasattr
from Products.Archetypes.debug import log_exc
from Products.CMFCore.utils import getToolByName
from OFS.Image import File, Image
from OFS.IOrderSupport import IOrderedContainer

try:
    from cPickle import dumps, UnpicklingError
except ImportError:
    from pickle import dumps, UnpicklingError

def getParent(item):
    inner = aq_inner(item)
    parent = aq_parent(inner)
    if parent is not None:
        return parent
    # A bug that we haven't been able to track yet
    if hasattr(inner, 'aq_parent'):
        return inner.aq_parent

def getReviewState(item):
    wt = getToolByName(item, 'portal_workflow')
    return wt.getInfoFor(item, 'review_state')

max_time = 1893466800.0 # Sane Unix Limit? 2030?
while True:
    try:
        time.mktime(time.gmtime(max_time))
    except OverflowError:
        max_time -= 1
    else:
        break

min_time = 21600.0 # mktime limit? 1970?
while True:
    try:
        time.mktime(time.gmtime(min_time))
        DateTime(min_time)
    except:
        min_time += 1
    else:
        break

class date:

    def __init__(self, name):
        self.name = name

    def __call__(self, item):
        if not shasattr(item, self.name):
            return DateTime()
        v = getattr(item, self.name)()
        if not v or not isinstance(v, DateTime):
            return DateTime()
        if float(v) > max_time:
            v = DateTime(max_time)
        if float(v) < min_time:
            v = DateTime(min_time)
        return v

def getOFSBlob(item):
    if not isinstance(item, (Image, File)):
        return None
    return str(item.data)

# Given a item and a method name, call the
# method to retrieve some value.
# If a tuple of names is provided, return the first name that returns data.
def getValue(item, attrs):
    if type(attrs) not in (ListType, TupleType):
        attrs = (attrs,)
    default = lambda : ''
    value = ''
    for attr_name in attrs:
        if isinstance(attr_name, basestring):
            if not shasattr(item, attr_name):
                continue
            attr = getattr(item, attr_name, default)
        else:
            # It's a function.
            func = attr_name
            def call(item=item):
                return func(item)
            attr = call
        __traceback_info__ = (item, attr_name, attr)
        try:
            if callable(attr):
                value = attr()
            else:
                value = attr
        except:
            log_exc('Failure extracting info %r:%r' % (item, attr_name))
            value = ''
        if value:
            return value
    return value

def make_blob(obj):
    if shasattr(obj, 'read'):
        return obj.read()
    if obj is not None:
        # What else?
        return str(obj)
    return ''

class InformationExtractor(object):

    implements(IInformationExtractor)

    attributes_map = {
        'id': ('getId',),
        'object_id': ('getId',),
        'uid': ('UID',),
        'title': ('Title',),
        'description': ('Description',),
        'keywords': ('Subject',),
        'url': ('absolute_url',),
        'credit': ('Contributors',),
        'body':('getText', 'CookedBody', 'getBody',),
        'review_state': (getReviewState,),
        'created': (date('created'),),
        'modified': (date('modified'),),
        'start_date': (date('start'),),
        'end_date': (date('end',),),
        'effective_date': (date('effective'),),
        'expires_date': (date('expires'),),
        'portal_type': ('getPortalTypeName', 'portal_type'),
        'meta_type': ('meta_type',),
        'blob': ('getFile', 'getImageAsFile', getOFSBlob),
        }

    formatters = {'blob': make_blob,
                  'body': make_blob}

    def __init__(self, context):
        self.context = context

    def extract(self):
        attr_map = self.attributes_map
        func = getValue
        fmts = self.formatters
        obj = self.context
        path = None
        if shasattr(obj, 'getPhysicalPath'):
            path = obj.getPhysicalPath()
        info = {}
        for attr, methods in attr_map.items():
            __traceback_info__ = (obj, path, attr, methods)
            value = func(obj, methods)
            f = fmts.get(attr, default_formatter)
            __traceback_info__ = (obj, path, attr, methods, f)
            value = f(value)
            # Catch Acquisition Wrappers. Will raise a TypeError on Zope 2.8.
            _ = getattr(value, '__reduce__', None)
            # If the above failed, try another way to see if the
            # object is acquisition wrapped
            if aq_base(value) is not value:
                raise TypeError(
                    'Found a acquisition wrapped object while '
                    'extracting attribute %s with '
                    'methods %s.' % (attr, repr(methods)))
            # If all else fails, and PICKLE_DEBUG is set, try to
            # pickle the object and let the exception pop all the way up.
            if PICKLE_DEBUG:
                # A negative value selects highest protocol available.
                _ = dumps(value, -1)
            info[attr] = value
        # Ordered if IOrderedSupport
        if shasattr(aq_base(obj), 'objectIds'):
            info['children'] = obj.objectIds()
        refs_adapter = IReferenceExtractor(obj, None)
        if refs_adapter is not None:
            # Should always find an adapter, we register one for '*'
            refs = refs_adapter.extract()
            if refs:
                assert not info.has_key('references'), (
                    "Conflict with key 'references'")
                # If we got this far, the object should have a UID()
                # method for sure.
                info['references'] = {'subject': obj.UID(),
                                      'predicates': refs}
        lr_adapter = ILocalRolesExtractor(obj, None)
        if lr_adapter is not None:
            # Should always find an adapter, we register one for '*'
            lr = lr_adapter.extract()
            if lr:
                assert not info.has_key('local_roles'), (
                    "Conflict with key 'local_roles'")
                info['local_roles'] = lr
        return info

class ReferenceExtractor(object):

    implements(IReferenceExtractor)

    def __init__(self, context):
        self.context = context

    def extract(self):
        item = self.context
        # XXX We are doing some sort of capabilities check here,
        # should eventually go away when we register the adapter only
        # for IReferenceable.
        if not shasattr(item, 'UID'):
            return None
        refs = {}
        rt = getToolByName(item, REFERENCE_CATALOG)
        for rname in rt.getRelationships(item):
            sID = item.UID()
            brains = rt._queryFor(sid=sID, relationship=rname)
            if brains:
                rels = {}
                for b in brains:
                    rels[b.targetUID] = None
                refs[rname] = rels.keys()
        parent = getParent(item)
        if parent is not None and shasattr(parent, 'UID'):
            refs['contained_into'] = (parent.UID(),)
        return refs

class LocalRolesExtractor(object):

    implements(ILocalRolesExtractor)

    def __init__(self, context):
        self.context = context

    def extract(self):
        item = self.context
        lr = {}
        for key, value in item.get_local_roles():
            lr[key] = Set(value)
        # Have to walk back up to the root and compute the sum of all
        # local role assignments.
        item = aq_parent(aq_inner(item))
        while item is not None:
            adapter = ILocalRolesExtractor(item, None)
            if adapter is None:
                break
            ilr = adapter.extract()
            for key, value in ilr.items():
                if not lr.has_key(key):
                    lr[key] = roles = Set()
                else:
                    roles = lr[key]
                roles.update(value)
            item = aq_parent(aq_inner(item))
        # Make the set of roles be a tuple.
        for key, value in lr.items():
            lr[key] = tuple(value)
        return lr

class DeploymentHandler(object):

    implements(IDeploymentHandler)

    def __init__(self, context):
        self.context = context

    def open(self, addr):
        self.handler = Client(addr)

    def close(self):
        self.handler.close()

    def deploy(self, info):
        self.handler.deploy(info)

class NullDeploymentHandler(object):

    implements(IDeploymentHandler)

    def __init__(self, context):
        self.context = context

    def open(self, addr): pass
    def close(self): pass
    def deploy(self, info): pass
