##############################################################################
#
# Enfold Stage Deployment - Remote Deployment of Staged Content
# Copyright (C) 2005 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##############################################################################
"""
$Id: subscriber.py 3390 2005-11-25 21:09:22Z sidnei $
"""

from urllib import unquote
from persistent import Persistent
from Acquisition import aq_base, aq_parent
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree
from BTrees.Length import Length
from Products.CMFCore.utils import getToolByName
from Products.Archetypes.utils import shasattr
from Products.EnSimpleStaging import event
from Products.EnSimpleStaging.content.stage import recurseVisit
from Products.StageDeployment.event import IDeployLabelEvent
from Products.StageDeployment.event import IPublishStageEvent
from Products.StageDeployment.event import IPublishLastChangedEvent
from Products.StageDeployment.utils import makeSocketAddress
from Products.StageDeployment.utils import getSocketAddress
from Products.StageDeployment.config import *
from Products.StageDeployment.interface import IInformationExtractor
from Products.StageDeployment.interface import IDeploymentHandler

class TargetInfo(Persistent):

    def __init__(self):
        self.last_index = Length()
        self.labels = IOBTree()

    def add(self, label):
        self.last_index.change(1)
        self.labels[self.last_index()] = label

    def last(self):
        return self.labels[self.last_index()]

# Given a set of items and attributes, extract
# the given attributes from each item, using a
# formatting function where it applies.
def extractInfo(context, items):
    info = []
    for item in items:
        extractor = IInformationExtractor(item)
        if extractor is None:
            # Should never happen, we register a default adapter for '*'
            continue
        logger.debug('Extracting %s' % '/'.join(item.getPhysicalPath()))
        info.append(extractor.extract())
        if getattr(item, '_p_changed', None) is None:
            if hasattr(item, '_p_deactivate'):
                item._p_deactivate()
    return info

def flushCache(conns):
    for c in conns:
        if hasattr(c, 'cacheMinimize'):
            c.cacheMinimize()
        if hasattr(c, 'cacheGC'):
            c.cacheGC()

def resolveObject(context, path, label=None, wt=None):
    if label is None:
        # Fetch current revision.
        # Content is inside the 'public' directory in a staging area.
        return context.unrestrictedTraverse('public/%s' % path)
    # Otherwise, fetch a specific revision as given by label.
    if wt is None:
        wt = getToolByName(context, 'portal_workspaces')
    vi = wt.getVersionInfoForPath(path=path, label=label,
                                  stage=context._getSourceStageName())
    if vi is None or vi.info is None:
        # Object didn't exist at the specified revision or it's
        # outside the Stage. Try and see if there's a current revision
        # that we can use.
        return context.unrestrictedTraverse('public/%s' % path)
    repo = wt._getVersionRepository()
    version = repo.getVersionOfResource(
        vi.info.history_id, vi.info.version_id)
    # Now work out the parent for this object.
    parts = filter(None, path.split('/'))
    parent_path = parts[:-1]
    if parent_path:
        # Parent is inside the stage.
        parent_path = '/'.join(parent_path)
        parent = resolveObject(context, parent_path, label=label, wt=wt)
    else:
        # Parent is not inside the stage, assume parent is the 'public' folder.
        parent = context.public
    return version.__of__(parent)

def deploymentInfo(context, mode, paths, conns, label=None):
    info = []
    items = []
    found = []
    for i, path in enumerate(paths):
        content = resolveObject(context, path, label=label)
        found.append(path)
        items.append(content)

        # Collect all affected connections
        conn = getattr(aq_base(content), '_p_jar', None)
        if conn is not None and not conns.has_key(conn):
            conns[conn] = None

        # Minimize cache and perform garbage collection.
        if not i % CACHE_THRESHOLD:
            flushCache(conns.keys())

    extracted = extractInfo(context, items)
    info.extend(zip(found, (mode,) * len(found), extracted))
    return info

def computeDeployment(stage, adds, updated, removes, label=None):
    info = []

    # We keep track of the ZODB connections used so that we can flush
    # them all at once.
    conns = {}

    # label must be a global label if informed.
    if label is not None:
        label = stage._globalLabel(label)

    # The operations order shouldn't really matter as we are not
    # making *incremental* changes, but snapshot-based changes, where
    # we have a delta instead of a set of operations performed.  All
    # paths should be ordered from longer to shorter, so that
    # containing folders are created correctly in case they hold any
    # metadata.
    for l in (adds, updated, removes):
        l.sort()
        l.reverse()

    # Removes don't need any special handling, just their paths.
    info.extend(zip(removes, ['remove'] * len(removes), [{}] * len(removes)))

    # For adds and updates we need to extract information to be passed
    # on to the deployment server.
    modes = (('add', adds),
             ('update', updated))
    for mode, paths in modes:
        info.extend(deploymentInfo(stage, mode, paths, conns, label=label))

    # Flush/Minimize all connection's caches.
    flushCache(conns.keys())
    return info

def paths(info):
    return ['/'.join(p) for p in info]

def getDeploymentTargetInfo(stage, target):
    if not shasattr(stage, '_deployment_info'):
        return
    d_info = getattr(stage, '_deployment_info')
    t_info = d_info.get(target)
    if t_info is None:
        return
    return t_info

def annotateDeployment(stage, target, label):
    if not shasattr(stage, '_deployment_info'):
        d_info = stage._deployment_info = OOBTree()
    else:
        d_info = stage._deployment_info
    t_info = d_info.get(target)
    if t_info is None:
        t_info = d_info[target] = TargetInfo()
    t_info.add(label)

def collectAll(stage):
    # No tag has been found, deploying for the first time it seems.
    base_url = stage.public.absolute_url(relative=1)
    if not base_url.endswith('/'):
        base_url = '%s/' % base_url
    removes = []
    updated = []
    adds = []
    def collect(obj):
        path = obj.absolute_url(relative=1).replace(base_url, '')
        adds.append(unquote(path))
    recurseVisit(stage.public, collect)
    return adds, updated, removes

def collectAllForLabel(stage, label):
    removes = []
    updated = []
    wt = getToolByName(stage, 'portal_workspaces')
    # label must be a global label
    label = stage._globalLabel(label)
    info_map = wt.getPath2VersionInfoMap(
        stage._getSourceStageName(), label)
    adds = ['/'.join(path) for path, history_id, version_id in info_map]
    adds.sort()
    return adds, updated, removes

# Event Subscribers
def publishedSubscriber(event):
    stage = event.obj
    addr = getSocketAddress(stage)
    try:
        delta = stage.getLastMovementsAndModifications()
        removes, adds = paths(delta['removes']), paths(delta['adds'])
        updated = paths(delta['modifications'])
    except ValueError:
        adds, updated, removes = collectAll(stage)
    deploy(stage, addr, adds, updated, removes)
    index, last_annotation = stage.getLastReversibleAnnotation()
    label = last_annotation['args']['label']
    # label must be a global label
    label = stage._globalLabel(label)
    annotateDeployment(stage, addr, label)

def deployLabelSubscriber(event):
    stage = event.obj
    label = event.label
    addr = event.target
    target_info = getDeploymentTargetInfo(stage, addr)
    if target_info is not None:
        last = target_info.last()
        delta = stage.getMovementsAndModifications(last, label)
        removes, adds = paths(delta['removes']), paths(delta['adds'])
        updated = paths(delta['modifications'])
    else:
        # No previous label published for this target, collect
        # existing paths at given label as 'adds'.
        adds, updated, removes = collectAllForLabel(stage, label)
    deploy(stage, addr, adds, updated, removes, label)
    annotateDeployment(stage, addr, label)

def deploy(stage, addr, adds, updated, removes, label=None):
    stage_path = '/'.join(stage.getPhysicalPath())
    logger.debug('Collecting deployment information for %s' % stage_path)
    info = computeDeployment(stage, adds, updated, removes, label=label)
    logger.debug('Finished collecting information for %s' % stage_path)
    if not info:
        return
    handler = IDeploymentHandler(stage)
    if isinstance(addr, basestring):
        addr = makeSocketAddress(addr)
    handler.open(addr)
    try:
        try:
            logger.debug('Starting deployment for %s' % stage_path)
            handler.deploy(info)
        except:
            logger.debug('Deployment failed for %s' % stage_path)
            raise
        else:
            logger.debug('Successfully deployed %s' % stage_path)
    finally:
        handler.close()

event.subscribe(IDeployLabelEvent, deployLabelSubscriber)
event.subscribe(IPublishLastChangedEvent, publishedSubscriber)
event.subscribe(IPublishStageEvent, publishedSubscriber)
